"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ts = require("typescript");
var path = require("path");
var crypto = require("crypto");
var prefix = crypto.randomBytes(256).toString('hex');
function hash(str) {
    var h = crypto.createHash('sha256');
    h.update(str);
    return h.digest('hex');
}
function apiTransformer(program) {
    function getSymbolId(symbol) {
        var uid = symbol.name + "#" + hash(prefix + symbol.id);
        return ts.createCall(ts.createIdentifier('Symbol.for'), [], [ts.createStringLiteral(uid)]);
    }
    function visitNodeAndChildren(node, program, context) {
        return ts.visitEachChild(visitNode(node, program), function (childNode) { return visitNodeAndChildren(childNode, program, context); }, context);
    }
    function visitNode(node, program) {
        var typeChecker = program.getTypeChecker();
        if (isCallExpression(node, typeChecker)) {
            var methodName = getCallExpressionName(node, typeChecker);
            if (methodName === 'InterfaceSymbol') {
                return visitSymbolNode(node, typeChecker);
            }
            else if (methodName === 'bind' || methodName === 'resolve' || methodName === 'resolveFactory') {
                return visitParametersNode(node, typeChecker);
            }
            else if (methodName === 'bindToImplements') {
                return bindToImplements(node, typeChecker);
            }
        }
        return node;
    }
    function visitSymbolNode(node, typeChecker) {
        if (!node.typeArguments) {
            throw new Error("No type arguments provided to `InterfaceSymbol<T>()`.");
        }
        var typeArgument = node.typeArguments[0];
        var type = typeChecker.getTypeFromTypeNode(typeArgument);
        var symbol = type.symbol;
        if (type.isClass() || !type.isClassOrInterface()) {
            throw new Error("The type provided is not an interface.");
        }
        return getSymbolId(symbol);
    }
    function getImports(root) {
        var fake = root;
        if (fake.imports && Array.isArray(fake.imports)) {
            return fake.imports;
        }
        else {
            return [];
        }
    }
    function getImportForType(typeChecker, imports, type) {
        for (var _i = 0, imports_1 = imports; _i < imports_1.length; _i++) {
            var node = imports_1[_i];
            if (ts.isImportDeclaration(node.parent)) {
                var importClause = node.parent.importClause;
                if (importClause && importClause.namedBindings && ts.isNamedImports(importClause.namedBindings)) {
                    var elements = importClause.namedBindings.elements;
                    for (var _a = 0, elements_1 = elements; _a < elements_1.length; _a++) {
                        var element = elements_1[_a];
                        if (ts.isImportSpecifier(element)) {
                            var importType = typeChecker.getTypeFromTypeNode(element.name);
                            var paramType = typeChecker.getTypeFromTypeNode(type);
                            if (importType.id === paramType.id) {
                                return element;
                            }
                        }
                    }
                }
            }
        }
        return undefined;
    }
    function getTypeExpression(node, sourceFile, typeChecker) {
        var expression;
        var isArray = false;
        var type = typeChecker.getTypeFromTypeNode(node);
        if (ts.isArrayTypeNode(node)) {
            var t = type;
            if (t.typeArguments && t.typeArguments.length > 0) {
                type = t.typeArguments[0];
                isArray = true;
            }
        }
        var symbol = type.symbol;
        if ((type.isClassOrInterface() && !type.isClass())) {
            expression = getSymbolId(symbol);
        }
        else {
            var imports = getImports(sourceFile);
            var importElement = getImportForType(typeChecker, imports, node);
            if (importElement) {
                expression = importElement.name;
            }
            else {
                expression = ts.createLiteral(symbol.name);
            }
        }
        if (isArray && expression) {
            expression = ts.createArrayLiteral([expression]);
        }
        return expression;
    }
    function getClassConstructorParameters(node, sourceFile, typeChecker) {
        var type = typeChecker.getTypeFromTypeNode(node);
        if (!type.symbol.members)
            return [];
        var params = [];
        type.symbol.members.forEach(function (val) {
            if (val.declarations.length === 0)
                return;
            var firstDeclaration = val.declarations[0];
            if (ts.isConstructorDeclaration(firstDeclaration)) {
                for (var _i = 0, _a = firstDeclaration.parameters; _i < _a.length; _i++) {
                    var param = _a[_i];
                    if (param.type) {
                        params.push(getTypeExpression(param.type, sourceFile, typeChecker));
                    }
                }
            }
        });
        return params;
    }
    function getClassImplementedInterfaces(node, sourceFile, typeChecker) {
        var type = typeChecker.getTypeFromTypeNode(node);
        var classDeclaration = type.symbol.valueDeclaration;
        if (!ts.isClassDeclaration(type.symbol.valueDeclaration))
            return [];
        if (!classDeclaration.heritageClauses)
            return [];
        var identifiers = [];
        for (var _i = 0, _a = classDeclaration.heritageClauses; _i < _a.length; _i++) {
            var clause = _a[_i];
            for (var _b = 0, _c = clause.types; _b < _c.length; _b++) {
                var type_1 = _c[_b];
                identifiers.push(getTypeExpression(type_1, sourceFile, typeChecker));
            }
        }
        return identifiers;
    }
    function visitParametersNode(node, typeChecker) {
        var signature = typeChecker.getResolvedSignature(node);
        if (!signature) {
            return node;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return node;
        }
        if (!ts.isMethodDeclaration(declaration) || !declaration.name) {
            return node;
        }
        var numArgs = declaration.name.getText() === 'bind' ? 2 : 1;
        if (node.arguments.length !== numArgs) {
            return node;
        }
        var args = [];
        for (var _i = 0, _a = node.arguments; _i < _a.length; _i++) {
            var val = _a[_i];
            args.push(val);
        }
        var newable = node.arguments[numArgs - 1];
        var params = getClassConstructorParameters(newable, node.getSourceFile(), typeChecker);
        args.push(ts.createArrayLiteral(params));
        node.arguments = ts.createNodeArray(args);
        return node;
    }
    function bindToImplements(node, typeChecker) {
        var signature = typeChecker.getResolvedSignature(node);
        if (!signature) {
            return node;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return node;
        }
        if (!ts.isMethodDeclaration(declaration) || !declaration.name) {
            return node;
        }
        if (node.arguments.length !== 1) {
            return node;
        }
        var args = [];
        for (var _i = 0, _a = node.arguments; _i < _a.length; _i++) {
            var val = _a[_i];
            args.push(val);
        }
        var sourceFile = node.getSourceFile();
        var newable = node.arguments[0];
        var params = getClassConstructorParameters(newable, sourceFile, typeChecker);
        args.push(ts.createArrayLiteral(params));
        var implementedInterfaces = getClassImplementedInterfaces(newable, sourceFile, typeChecker);
        args.push(ts.createArrayLiteral(implementedInterfaces));
        node.arguments = ts.createNodeArray(args);
        return node;
    }
    var apiTs = path.dirname(__dirname).replace(/\\/g, "/");
    function getCallExpressionName(node, typeChecker) {
        if (!ts.isCallExpression(node)) {
            return undefined;
        }
        var signature = typeChecker.getResolvedSignature(node);
        if (typeof signature === 'undefined') {
            return undefined;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return undefined;
        }
        var dirname = path.dirname(declaration.getSourceFile().fileName).replace(/\\/g, "/");
        if (apiTs !== dirname && dirname.indexOf(apiTs) !== 0) {
            return undefined;
        }
        if ((ts.isMethodDeclaration(declaration) || ts.isFunctionDeclaration(declaration)) && !!declaration.name) {
            return declaration.name.getText();
        }
        return undefined;
    }
    function isCallExpression(node, typeChecker) {
        if (node.kind !== ts.SyntaxKind.CallExpression) {
            return false;
        }
        var signature = typeChecker.getResolvedSignature(node);
        if (typeof signature === 'undefined') {
            return false;
        }
        var declaration = signature.declaration;
        if (!declaration) {
            return false;
        }
        return (ts.isMethodDeclaration(declaration)
            || ts.isFunctionDeclaration(declaration)) && !!declaration.name;
    }
    return function (context) { return function (file) { return visitNodeAndChildren(file, program, context); }; };
}
exports.default = apiTransformer;
//# sourceMappingURL=transformer.js.map